/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.Utilities;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * This class provides information about the file system structure of different card types.
 * The information is backed by JSON configuration files that are provided as app resources.
 *
 * @author André Szdzuy
 */
public class CardFileSystem {

    private JSONObject structure = null;

    private static CardFileSystem egkInstance = null;
    private static CardFileSystem hbaInstance = null;
    private static CardFileSystem smcbInstance = null;

    /**
     * Get an instance representing the structure of the eGK (Gesundheitskarte).
     *
     * @return the instance
     * @throws IOException if a file reading error occurred
     * @throws JSONException if JSON parsing error occurred
     */
    public static CardFileSystem getEGKInstance() throws IOException, JSONException {
        if (egkInstance == null) {
            egkInstance = new CardFileSystem("egk.json");
        }

        return egkInstance;
    }

    /**
     * Get an instance representing the structure of the HBA (Heilberufeausweis).
     *
     * @return the instance
     * @throws IOException if a file reading error occurred
     * @throws JSONException if a JSON parsing error occurred
     */
    public static CardFileSystem getHBAInstance() throws IOException, JSONException {
        if (hbaInstance == null) {
            hbaInstance = new CardFileSystem("hba.json");
        }

        return hbaInstance;
    }

    /**
     * Get an instance representing the structure of the SMC-B (Sicherheitsmodulkarte für Institutionen im Gesundheitswesen).
     *
     * @return the instance
     * @throws IOException if a file reading error occurred
     * @throws JSONException if a JSON parsing error occurred
     */
    public static CardFileSystem getSMCBInstance() throws IOException, JSONException {
        if (smcbInstance == null) {
            smcbInstance = new CardFileSystem("smcb.json");
        }

        return smcbInstance;
    }

    /**
     * Internal constructor.
     *
     * @throws IOException if a file reading error occurred
     * @throws JSONException if a JSON parsing error occurred
     */
    private CardFileSystem(String jsonFile) throws IOException, JSONException {
        ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
        InputStream is = classLoader.getResourceAsStream(jsonFile);
        String s = Utilities.readTextFromStream(is);
        this.structure = new JSONObject(s);
    }

    /**
     * Get the name of the DF (Dedicated File) identified by the given AID.
     *
     * @param aid application identifier of the DF
     * @return name of the DF
     * @throws JSONException if a JSON parsing error occurred
     */
    public String getDF(String aid) throws JSONException {
        if (aid == null) {
            return null;
        }
        JSONObject o = getCompleteDF(aid);
        return o.getString("name");
    }

    /**
     * Get the name of the EF (Elementery File) identified by the given FID.
     *
     * @param aid application identifier of the parent DF
     * @param fid file identifier
     * @return name of the EF
     * @throws JSONException if a JSON parsing error occurred
     */
    public String getEF(String aid, String fid) throws JSONException {
        if (aid == null || fid == null) {
            return null;
        }

        JSONObject o = getCompleteDF(aid);
        o = o.getJSONObject("files");
        return o.getString(fid);
    }

    /**
     * @return all AIDs that exist on the card
     */
    public Iterator<String> getAllAids() {
        List<String> result = new ArrayList<String>();
        String root = structure.optString("aid");
        if (root != null && !"".equals(root)) {
            result.add(root);
        }

        JSONArray folders = structure.optJSONArray("folders");
        if (folders != null) {
            for (int i = 0; i < folders.length(); i++) {
                JSONObject item = folders.optJSONObject(i);
                if (item != null) {
                    String aid = item.optString("aid");
                    if (aid != null && !"".equals(aid)) {
                        result.add(aid);
                    }
                }
            }
        }

        return result.iterator();
    }

    /**
     * Get the FIDs of all children (EFs) of the DF identified by the given AID.
     *
     * @param aid application identifier of the parent DF
     * @return the FIDs of the children
     * @throws JSONException if a JSON parsing error occurred
     */
    public Iterator<String> getFids(String aid) throws JSONException {
        JSONObject o = getCompleteDF(aid);
        o = o.optJSONObject("files");
        if (o == null) {
            return new ArrayList<String>().iterator();  // empty iterator
        } else {
            return o.keys();
        }
    }

    /**
     * @return JSON representation of the file system structure
     */
    public String toString() {
        return structure.toString();
    }

    private JSONObject getCompleteDF(String aid) throws JSONException {
        String root = structure.getString("aid");
        if (aid.equals(root)) {
            return structure;
        }
        JSONArray folders = structure.getJSONArray("folders");
        for (int i = 0; i < folders.length(); i++) {
            JSONObject item = folders.getJSONObject(i);
            if (aid.equals(item.getString("aid"))) {
                return item;
            }
        }
        return null;
    }

}
