/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.Utilities;

import de.gematik.security.iso7816.FCP;
import org.json.JSONException;
import org.spongycastle.asn1.ASN1InputStream;
import org.spongycastle.asn1.util.ASN1Dump;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;

/**
 * This class holds the data of a file (DF or EF) on the card.
 *
 * @author Ronald Koenig
 * @author André Szdzuy
 */
public class CardItemFile {

    private static final String TAG = CardItemFile.class.getName();

    private String name;
    private FCP fcp;
    private FCP parent;
    private ArrayList<byte[]> data;

    public CardItemFile(String name, byte[] data) {
        this.name = name;
        this.data = new ArrayList<byte[]>();
        this.data.add(data);
    }

    public CardItemFile(String name, ArrayList<byte[]> data) {
        this.name = name;
        this.data = data;
    }

    /**
     * Constructor. Assumes EGK file system.
     *
     * @param fcp metadata of the file
     * @param parent metadata of the parent file
     * @param data data of the file
     * @throws IOException if a file reading error occurred
     * @throws JSONException if JSON parsing error occurred
     */
    public CardItemFile(FCP fcp, FCP parent, ArrayList<byte[]> data) throws IOException, JSONException {
        this(CardFileSystem.getEGKInstance(), fcp, parent, data);
    }

    /**
     * Constructor.
     *
     * @param cfs card file system (EGK or HBA)
     * @param fcp metadata of the file
     * @param parent metadata of the parent file
     * @param data data of the file
     * @throws IOException if a file reading error occurred
     * @throws JSONException if JSON parsing error occurred
     */
    public CardItemFile(CardFileSystem cfs, FCP fcp, FCP parent, ArrayList<byte[]> data) throws IOException, JSONException {
        if (fcp.getDfName() != null) {
            String aid = Utilities.ByteArrayToHexString(fcp.getDfName());
            name = cfs.getDF(aid);
            if (name == null) {
                name = aid;
            }
        } else {
            String aid = Utilities.ByteArrayToHexString(parent.getDfName());
            String fid = Utilities.ByteArrayToHexString(fcp.getFileIdentifier());
            name = cfs.getEF(aid, fid);
            if (name == null) {
                name = fid;
            }
        }
        this.fcp = fcp;
        this.parent = parent;
        this.data = data;
    }

    /**
     * @return a string representation of the file (its name)
     */
    public String toString() {
        return fcp.getDfName() != null ? name : "   " + name;
    }

    /**
     * @return name of the file
     */
    public String getName() {
        return name;
    }

    /**
     * Set name.
     *
     * @param name name of the file
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return metadata of the file
     */
    public FCP getFcp() {
        return fcp;
    }

    /**
     * @return metadata of the parent file
     */
    public FCP getParent() {
        return parent;
    }

    /**
     * @return the data of the file
     */
    public ArrayList<byte[]> getData() {
        return data;
    }

    public String unzipXML() throws IOException {
        String xmlDoc;
        if ("EF.VD".equals(name)) {
            int index = 0;
            //Read offsets
            ByteBuffer byteBuffer = ByteBuffer.wrap(data.get(0));
            short[] offset = new short[4];
            byteBuffer.asShortBuffer().get(offset, 0, 4);
            xmlDoc = Utilities.unzip(data.get(0), offset[index * 2], offset[index * 2 + 1] - offset[index * 2] + 1);
        } else {    // "EF.PD", "EF.GVD", "EF.NFD"
            //Read length
            ByteBuffer byteBuffer = ByteBuffer.wrap(data.get(0));
            int length = byteBuffer.asShortBuffer().get();
            xmlDoc = Utilities.unzip(data.get(0), 2, length);
        }
        return xmlDoc;
    }

    /**
     * @return a nicely formatted representation of the file's data
     * @throws CertificateException if a certificate error occurred
     */
    public String displayData(boolean parseXML) throws CertificateException, ParserConfigurationException, SAXException, IOException {
        String result = "no data";
        if (data != null) {
            if ("EF.PD".equals(name) || "EF.VD".equals(name) || "EF.GVD".equals(name) || "EF.NFD".equals(name)) {
                String xmlDoc = unzipXML();
                if (parseXML) {
                    result = Utilities.displayXmlElements(xmlDoc);
                } else {
                    StringBuilder stringBuilder = new StringBuilder();
                    stringBuilder.append("data:text/xml;charset=UTF-8,");
                    stringBuilder.append(xmlDoc);
                    result = stringBuilder.toString();
                }
            } else if ("EF.AUT".equals(name) || "EF.AUTN".equals(name) || "EF.ENC".equals(name) || "EF.ENCV".equals(name)
                    || "EF.C.HP.QES.R2048".equals(name)) {
                X509Certificate x509Cert = Utilities.getX509Certificate(data.get(0));
                if (x509Cert != null) {
                    result = x509Cert.toString();
                }
            } else if ("EF.ATR".equals(name) || "EF.CardAccess".equals(name) || "EF.GDO".equals(name) || "EF.VERSION2".equals(name)
                    || "EF.C.CA_eGK.CS.E256".equals(name) || "EF.C.eGK.AUT_CVC.E256".equals(name)
                    || "EF.C.CA_HPC.CS.R2048".equals(name) || "EF.C.CA_HPC.CS.E256".equals(name) || "EF.C.HPC.AUTR_CVC.R2048".equals(name)
                    || "EF.C.HPC.AUTR_CVC.E256".equals(name) || "EF.C.HPC.AUTD_SUK_CVC.E256".equals(name)) {
                StringBuilder stringBuilder = new StringBuilder();
                ASN1InputStream asn1InputStream = new ASN1InputStream(new ByteArrayInputStream(data.get(0)));
                while (asn1InputStream.available() > 0) {
                    stringBuilder.append(ASN1Dump.dumpAsString(asn1InputStream.readObject()));
                }
                result = stringBuilder.toString();
            } else if ("EF.Logging".equals(name)) {
                StringBuilder stringBuilder = new StringBuilder();
                for (byte[] record : data) {
                    if (record.length > 0) {
                        stringBuilder.append(new String(record, "ISO-8859-1"));
                        stringBuilder.append("\n");
                    }
                }
                if (stringBuilder.length() > 0) {
                    result = stringBuilder.toString();
                }
            } else {
                StringBuilder stringBuilder = new StringBuilder();
                for (byte[] record : data) {
                    if (record.length > 0) {
                        stringBuilder.append(Utilities.ByteArrayToSpacedHexString(record));
                        stringBuilder.append("\n");
                    }
                }
                if (stringBuilder.length() > 0) {
                    result = stringBuilder.toString();
                }
            }
        }

        return result;
    }

}



