/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.Utilities;

import com.google.api.client.http.HttpHeaders;
import com.google.api.client.http.HttpRequestFactory;
import com.google.api.client.http.HttpTransport;
import com.google.api.client.http.javanet.NetHttpTransport;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONObject;

import javax.smartcardio.*;
import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * This class is a smartcardio CardChannel that executes the APDU commands on a remote card service.
 *
 * @author André Szdzuy
 */
public class RemoteCardChannel extends CardChannel {
    private static final Log LOG = LogFactory.getLog(RemoteCardChannel.class);

    private HttpTransport httpTransport = null;
    private HttpRequestFactory httpRequestFactory = null;
    private HttpHeaders headers = null;
    private String serverURL = null;

    /**
     * Constructs a RemoteCardChannel for use with the Jersey-based remote card service.
     *
     * @param serverURL URL of remote card service
     * @param authorizationHeader authorization header
     * @throws CardException if an error occurred
     */
    public RemoteCardChannel(String serverURL, String authorizationHeader) throws CardException {
        this.httpTransport = new NetHttpTransport();
        this.httpRequestFactory = httpTransport.createRequestFactory();

        if (authorizationHeader != null) {
            this.headers = new HttpHeaders();
            this.headers.setAuthorization(authorizationHeader);
        }

        try {
            byte[] initialResponse =
                    Utilities.sendBinaryHttpRequest(httpRequestFactory, serverURL, new byte[] {}, headers);
            JSONObject jo = new JSONObject(new String(initialResponse, "UTF-8"));
            String sessionID = jo.getString("sessionID");
            String creationTime = jo.getString("creationTime");
            LOG.info("Started session with sessionID = " + sessionID + " (creationTime = " + creationTime + ")");
            this.serverURL = serverURL + "/" + sessionID;
        } catch (IOException e) {
            LOG.error(e);
            throw new CardException(e);
        }
    }

    /**
     * Constructs a RemoteCardChannel for use with a down-stripped NanoHTTP-based remote card service.
     * Just for testing without the need for sessions and authorizationHeader.
     *
     * @param serverURL URL of remote card service
     * @throws CardException if an error occurred
     */
    public RemoteCardChannel(String serverURL) throws CardException {
        this.httpTransport = new NetHttpTransport();
        this.httpRequestFactory = httpTransport.createRequestFactory();
        this.serverURL = serverURL;
    }

    @Override
    public Card getCard() {
        return null;
    }

    @Override
    public int getChannelNumber() {
        return 0;
    }

    @Override
    public ResponseAPDU transmit(CommandAPDU commandAPDU) throws CardException {
        try {
            byte[] apduBytes = commandAPDU.getBytes();
            byte[] responseContent =
                Utilities.sendBinaryHttpRequest(httpRequestFactory, serverURL + "/executeCommand", apduBytes, headers);
            return new ResponseAPDU(responseContent);
        } catch (IOException e) {
            LOG.error(e);
            throw new CardException(e);
        }
    }

    @Override
    public int transmit(ByteBuffer byteBuffer, ByteBuffer byteBuffer1) throws CardException {
        return 0;
    }

    @Override
    public void close() throws CardException {

    }
}
