/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.api;

import de.gematik.security.Utilities.CardFileSystem;
import de.gematik.security.Utilities.CardItemFile;
import de.gematik.security.iso7816.FCP;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import de.gematik.security.iso9564.PinBlock;
import de.gematik.security.iso9564.PinBlockException;
import org.json.JSONException;

import javax.smartcardio.CardException;
import java.io.IOException;
import java.util.ArrayList;

/**
 * @author André Szdzuy
 */
public interface Iso7816Commands {

    /**
     * Selects root file
     * <p/>
     *
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectADF() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects root file or next dedicated file with attribute application identifier
     * <p/>
     * Selects next occurrence of dedicated file with attribute application identifier
     * if called repeatedly and nextOccurrence is true
     *
     * @param nextOccurrence if true next file occurrence else first file occurrence
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectADF(boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects first occurrence of dedicated file by applicationIdentifier
     * <p/>
     *
     * @param applicationIdentifier the applicationIdentifier of the dedicated file
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectADF(String applicationIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects dedicated file by applicationIdentifier
     * <p/>
     * Selects next occurrence of dedicated file if called repeatedly and nextOccurrence is true
     *
     * @param applicationIdentifier the applicationIdentifier of the dedicated file
     * @param nextOccurrence        if true next file occurrence else first file occurrence
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectADF(String applicationIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects parent dedicated file
     * <p/>
     *
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectDF() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects first occurrence of dedicated file by fileIdentifier
     * <p/>
     * Not applicable for eGK because there exist no DFs with a fileIdentifier.
     *
     * @param fileIdentifier the fileIdentifier of the dedicated file
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectDF(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects dedicated file by fileIdentifier
     * <p/>
     * Selects next occurrence of dedicated file if called repeatedly and nextOccurrence is true.
     * Not applicable for eGK because there exist no DFs with a fileIdentifier.
     *
     * @param fileIdentifier the fileIdentifier of the dedicated file
     * @param nextOccurrence if true next file occurrence else first file occurrence
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectDF(String fileIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects elementary file by fileIdentifier
     * <p/>
     *
     * @param fileIdentifier the fileIdentifier of the elementary file
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void selectEF(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects root file
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectADFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects root file or next dedicated file with attribute application identifier
     * <p/>
     * Selects next occurrence of dedicated file with attribute application identifier
     * if called repeatedly and nextOccurrence is true
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param nextOccurrence if true next file occurrence else first file occurrence
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectADFAndReturnFCP(boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects first occurrence of dedicated file by applicationIdentifier
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param applicationIdentifier the applicationIdentifier of the dedicated file
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectADFAndReturnFCP(String applicationIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects dedicated file by applicationIdentifier
     * <p/>
     * Selects next occurrence of dedicated file if called repeatedly and nextOccurrence is true
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param applicationIdentifier the applicationIdentifier of the dedicated file
     * @param nextOccurrence        if true next file occurrence else first file occurrence
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectADFAndReturnFCP(String applicationIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects parent dedicated file
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectDFAndReturnFCP() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects first occurrence of dedicated file by fileIdentifier
     * <p/>
     * Not applicable for eGK because there exist no DFs with a fileIdentifier.
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param fileIdentifier the fileIdentifier of the dedicated file
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectDFAndReturnFCP(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects dedicated file by fileIdentifier
     * <p/>
     * Selects next occurrence of dedicated file if called repeatedly and nextOccurrence is true
     * Not applicable for eGK because there exist no DFs with a fileIdentifier.
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param fileIdentifier the fileIdentifier of the dedicated file
     * @param nextOccurrence if true next file occurrence else first file occurrence
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectDFAndReturnFCP(String fileIdentifier, boolean nextOccurrence) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Selects elementary file by fileIdentifier
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param fileIdentifier the fileIdentifier of the elementary file
     * @return file control parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    FCP selectEFAndReturnFCP(String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Reads bytes from the current file starting at offset 0 until end of file is reached.
     * The bytes actually read are returned as byte array.
     *
     * @return number of bytes actually read
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    byte[] read() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Reads bytes from the current file starting at offset.
     * Stops reading if either len number of bytes are read or end of file is reached.
     * The bytes actually read are returned as byte array.
     *
     * @param offset starting offset within file
     * @param len    number of bytes to read
     * @return array of bytes actually read
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    byte[] read(int offset, int len) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Reads record indicated by recordNumber from the current file until end of record is reached.
     * The bytes actually read are returned as byte array.
     *
     * @param recordNumber the number of the record to be read
     * @return array of bytes actually read
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    byte[] readRecord(int recordNumber) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Reads all records from the current file until end of file is reached.
     * The bytes actually read are returned as a list of byte arrays (one byte array per record).
     *
     * @return bytes actually read
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    ArrayList<byte[]> readAllRecords() throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Reads a complete elementary file. Assumes EGK file system.
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param parent         parent dedicated file
     * @param fileIdentifier the fileIdentifier of the elementary file
     * @return a CardItemFile
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    CardItemFile readEF(FCP parent, String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException, IOException, JSONException;

    /**
     * Reads a complete elementary file.
     * <p/>
     * Should only be used in contexts where Extended Length Apdus are supported!
     *
     * @param cfs            card file system (EGK or HBA)
     * @param parent         parent dedicated file
     * @param fileIdentifier the fileIdentifier of the elementary file
     * @return a CardItemFile
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    CardItemFile readEF(CardFileSystem cfs, FCP parent, String fileIdentifier) throws Iso7816UnsuccessfulExecutionException, CardException, IOException, JSONException;

    /**
     * Writes data to a transparent file (UPDATE BINARY). Always starts at offset 0.
     *
     * @param data the data to write
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void update(byte[] data) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
         * Verifies secret
         * <p/>
         * Verifies the secret and sets the security status of the card according to the verification result.
         * Automatically finds out the type of the card (EGK, HBA or SMC-B) and uses the corresponding secretReference.
         *
         * @param verificationData integer string the secret is verified against
         * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
         *                                               (SW is stored in detailed message)
         * @throws CardException                         any Card IO exception
         * @throws PinBlockException                     if PIN format is invalid
         */
    void verify(String verificationData) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException;

    /**
     * Verifies secret
     * <p/>
     * Verifies the secret and sets the security status of the card according to the verification result
     *
     * @param secretReference  the reference to the secret which shall be verified
     * @param verificationData integer string the secret is verified against
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     * @throws PinBlockException                     if PIN format is invalid
     */
    void verify(int secretReference, String verificationData) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException;

    /**
     * Verifies secret
     * <p/>
     * Verifies the secret and sets the security status of the card according to the verification result
     *
     * @param secretReference  the reference to the secret which shall be verified
     * @param verificationData PinBlock the secret is verified against
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void verify(int secretReference, PinBlock verificationData) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Unlocks the card if verification is blocked due to too many failed verify attempts or if card is in transportStatus.
     * Automatically finds out the type of the card (EGK, HBA or SMC-B) and uses the corresponding secretReference.
     *
     * @param puk       optional PUK, can also be null
     * @param newSecret optional new PIN, can also be null
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     * @throws PinBlockException                     if PIN or PUK format is invalid
     */
    void unlock(String puk, String newSecret) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException;

    /**
     * Unlocks the card if verification is blocked due to too many failed verify attempts or if card is in transportStatus.
     *
     * @param secretReference the reference to the blocked secret
     * @param puk             optional PUK, can also be null
     * @param newSecret       optional new PIN, can also be null
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     * @throws PinBlockException                     if PIN or PUK format is invalid
     */
    void unlock(int secretReference, String puk, String newSecret) throws Iso7816UnsuccessfulExecutionException, CardException, PinBlockException;

    /**
     * Unlocks the card if verification is blocked due to too many failed verify attempts or if card is in transportStatus.
     *
     * @param secretReference the reference to the blocked secret
     * @param puk             optional PUK, can also be null
     * @param newSecret       optional new PIN, can also be null
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     * @throws PinBlockException                     if PIN or PUK format is invalid
     */
    void unlock(int secretReference, PinBlock puk, PinBlock newSecret) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Executes an MSE (MANAGE SECURITY ENVIRONMENT) command to set a key for use by following commands.
     *
     * @param prefix prefix for command data
     * @param keyRef key reference to set
     * @param algId  algorithm to set
     * @param p1     value of p1 command apdu parameter
     * @param p2     value of p2 command apdu parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void manageSE(String prefix, String keyRef, String algId, int p1, int p2) throws Iso7816UnsuccessfulExecutionException, CardException;

    /**
     * Executes an MSE (MANAGE SECURITY ENVIRONMENT) command to set a key for use by following commands.
     *
     * @param prefix prefix for command data
     * @param keyRef key reference to set
     * @param p1     value of p1 command apdu parameter
     * @param p2     value of p2 command apdu parameter
     * @throws Iso7816UnsuccessfulExecutionException if the card commands return an error
     *                                               (SW is stored in detailed message)
     * @throws CardException                         any Card IO exception
     */
    void manageSE(String prefix, String keyRef, int p1, int p2) throws Iso7816UnsuccessfulExecutionException, CardException;

}
