/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.card2server;

import de.gematik.security.Utilities.Utilities;
import de.gematik.security.crypto.AmAESCrypto;
import de.gematik.security.crypto.AmCryptoException;
import de.gematik.security.crypto.AmCryptoProvider;
import de.gematik.security.tr03110.KeyDerivationFunction;
import de.gematik.security.tr03110.SecureMessaging;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.smartcardio.*;
import java.nio.ByteBuffer;
import java.util.Arrays;

/**
 * A CardChannel implementation that represents the "Gegenstelle" in the Card2Server descriptions.
 *
 * @author André Szdzuy
 */
public class ServerChannel extends CardChannel {
    private static final Log LOG = LogFactory.getLog(ServerChannel.class);

    private byte[] rnd16_svr = new byte[16];
    private byte[] iccsn8_svr = new byte[8];
    private byte[] kd64_svr = new byte[64];
    private byte[] rnd16_egk = new byte[16];
    private byte[] iccsn8_egk = new byte[8];
    private byte[] kd64_egk = new byte[64];

    // MF/SK.VSD.AES256:
    private byte[] encKey = null;
    private byte[] macKey = null;

    private SecureMessaging secureMessaging = null;

    public ServerChannel(byte[] encKey, byte[] macKey) {
        this.encKey = encKey;
        this.macKey = macKey;
    }

    @Override
    public Card getCard() {
        return null;
    }

    @Override
    public int getChannelNumber() {
        return 0;
    }

    @Override
    public ResponseAPDU transmit(CommandAPDU commandAPDU) throws CardException {
        Utilities.logCommandAPDU(commandAPDU);
        byte[] result = null;

        if (commandAPDU.getINS() == 0x88) {     // INTERNAL AUTHENTICATE
            // fill rnd16_svr
            for (int i = 0; i < rnd16_svr.length; i++) {
                rnd16_svr[i] = (byte) 0x00;     // actually we need random values here, but 0 will also do for the moment
            }
            // fill iccsn8_svr
            for (int i = 0; i < iccsn8_svr.length; i++) {
                iccsn8_svr[i] = (byte) 0x00;     // actually we need random values here, but 0 will also do for the moment
            }
            // fill kd64_svr
            for (int i = 0; i < kd64_svr.length; i++) {
                kd64_svr[i] = (byte) 0x00;     // actually we need random values here, but 0 will also do for the moment
            }
            // get rnd16_egk and iccsn8_egk from commandAPDU.getData()
            byte[] iaInput = commandAPDU.getData();
            System.arraycopy(iaInput, 0, rnd16_egk, 0, 16);
            System.arraycopy(iaInput, 16, iccsn8_egk, 0, 8);

            // concatenate rnd16_svr, iccsn8_svr, rnd16_egk, iccsn8_egk, kd64_svr
            // (rnd16_egk and iccsn8_egk can be taken from commandAPDU.getData())
            byte[] concatenation = new byte[112];
            System.arraycopy(rnd16_svr, 0, concatenation, 0, 16);
            System.arraycopy(iccsn8_svr, 0, concatenation, 16, 8);
            System.arraycopy(rnd16_egk, 0, concatenation, 24, 16);
            System.arraycopy(iccsn8_egk, 0, concatenation, 40, 8);
            System.arraycopy(kd64_svr, 0, concatenation, 48, 64);
            LOG.info("concatenation: " + Utilities.ByteArrayToSpacedHexString(concatenation));

            try {
                // encrypt concatenation with MF/SK.VSD.AES256
                AmCryptoProvider encCrypto = new AmAESCrypto();
                byte[] encryptedConcatenation = encCrypto.encrypt(encKey, concatenation);

                // encryption alternative with JCE classes instead BouncyCastle classes:
//                SecretKeySpec skeySpec = new SecretKeySpec(encKey, "AES");
//                Cipher encryptor = Cipher.getInstance("AES/CBC/NoPadding");
//                encryptor.init(Cipher.ENCRYPT_MODE, skeySpec, new IvParameterSpec(new byte[16]));
//                byte[] encryptedConcatenation = encryptor.doFinal(concatenation);

                LOG.info("encryptedConcatenation: " + Utilities.ByteArrayToSpacedHexString(encryptedConcatenation));

                // compute MAC on the encrypted concatenation
                AmCryptoProvider macCrypto = new AmAESCrypto();
                byte[] mac = macCrypto.getMAC(macKey, encryptedConcatenation, true);
                LOG.info("mac: " + Utilities.ByteArrayToSpacedHexString(mac));

                result = new byte[122];
                System.arraycopy(encryptedConcatenation, 0, result, 0, encryptedConcatenation.length);
                System.arraycopy(mac, 0, result, encryptedConcatenation.length, mac.length);
            } catch (AmCryptoException e) {
                LOG.error(e);
                throw new CardException(e);
            }

            // append SW bytes for OK
            result[120] = (byte) 0x90;
            result[121] = (byte) 0x00;
        } else if (commandAPDU.getINS() == 0x82) {     // EXTERNAL AUTHENTICATE
            byte[] maOutput = commandAPDU.getData();
            byte[] encryptedConcatenation = new byte[112];
            byte[] mac = new byte[8];
            System.arraycopy(maOutput, 0, encryptedConcatenation, 0, encryptedConcatenation.length);
            System.arraycopy(maOutput, encryptedConcatenation.length, mac, 0, mac.length);
            LOG.info("encryptedConcatenation: " + Utilities.ByteArrayToSpacedHexString(encryptedConcatenation));
            LOG.info("mac: " + Utilities.ByteArrayToSpacedHexString(mac));

            try {
                // compute MAC on the encrypted concatenation and compare it with the given value
                AmCryptoProvider macCrypto = new AmAESCrypto();
                byte[] computedMac = macCrypto.getMAC(macKey, encryptedConcatenation, true);
                if (!Arrays.equals(mac, computedMac)) {
                    // return AuthenticationFailure
                    result = new byte[] {(byte) 0x63, (byte) 0x00};
                } else {
                    // decrypt concatenation with MF/SK.VSD.AES256
                    AmCryptoProvider encCrypto = new AmAESCrypto();
                    byte[] concatenation = encCrypto.decrypt(encKey, encryptedConcatenation);
                    LOG.info("concatenation: " + Utilities.ByteArrayToSpacedHexString(concatenation));

                    // split concatenation in rnd16_egk, iccsn8_egk, rnd16_svr, iccsn8_svr, kd64_egk
                    byte[] rnd16_egk_toCheck = new byte[16];
                    byte[] iccsn8_egk_toCheck = new byte[8];
                    byte[] rnd16_svr_toCheck = new byte[16];
                    byte[] iccsn8_svr_toCheck = new byte[8];
                    System.arraycopy(concatenation, 0, rnd16_egk_toCheck, 0, 16);
                    System.arraycopy(concatenation, 16, iccsn8_egk_toCheck, 0, 8);
                    System.arraycopy(concatenation, 24, rnd16_svr_toCheck, 0, 16);
                    System.arraycopy(concatenation, 40, iccsn8_svr_toCheck, 0, 8);
                    System.arraycopy(concatenation, 48, kd64_egk, 0, 64);

                    // verify rnd16_egk, iccsn8_egk, rnd16_svr, iccsn8_svr
                    if (!Arrays.equals(rnd16_egk, rnd16_egk_toCheck)
                            || !Arrays.equals(iccsn8_egk, iccsn8_egk_toCheck)
                            || !Arrays.equals(rnd16_svr, rnd16_svr_toCheck)
                            || !Arrays.equals(iccsn8_svr, iccsn8_svr_toCheck))
                    {
                        // return AuthenticationFailure
                        result = new byte[] {(byte) 0x63, (byte) 0x00};
                    }
                }
            } catch (AmCryptoException e) {
                LOG.error(e);
                throw new CardException(e);
            }

            // authentication successful, now derive key for trusted channel and init SecureMessaging

            byte[] keyDerivationData = new byte[64];

            // kd64_svr XOR kd64_egk
            for (int i = 0; i < 64; i++) {
                keyDerivationData[i] = (byte) (kd64_svr[i] ^ kd64_egk[i]);
            }

            byte[] kenc = KeyDerivationFunction.getAES256Key(keyDerivationData, KeyDerivationFunction.Mode.ENC);
            byte[] kmac = KeyDerivationFunction.getAES256Key(keyDerivationData, KeyDerivationFunction.Mode.MAC);

            AmCryptoProvider crypto = new AmAESCrypto();
            secureMessaging = new SecureMessaging(crypto, kenc, kmac, new byte[crypto.getBlockSize()]);

            if (result == null) {
                // return just OK
                result = new byte[] {(byte) 0x90, (byte) 0x00};
            }
        } else {
            // return just OK
            result = new byte[] {(byte) 0x90, (byte) 0x00};
        }

        return new ResponseAPDU(result);
    }

    @Override
    public int transmit(ByteBuffer byteBuffer, ByteBuffer byteBuffer1) throws CardException {
        return 0;
    }

    @Override
    public void close() throws CardException {

    }

    public SecureMessaging getSecureMessaging() {
        return secureMessaging;
    }

}
