/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.crypto;

import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;

import javax.smartcardio.CardException;
import java.io.IOException;
import java.security.*;
import java.security.spec.AlgorithmParameterSpec;
import java.util.Arrays;
import java.util.HashMap;

/**
 * Implementation of the JCA/JCE signature interface that uses card operations to compute
 * signatures.
 *
 * @author André Szdzuy
 */
public class CardBasedSignature extends Signature {

    private String algorithm;
    private HashMap<String, Object> parameters = new HashMap<String, Object>();
    private AlgorithmParameterSpec aps = null;

    private int dfSpecificPrivateKeyRef = -1;
    private HealthcardContext card = null;
    private byte[] input = null;

    protected CardBasedSignature(String algorithm) throws NoSuchAlgorithmException {
        super(algorithm);
        if (!"SHA256withRSA".equals(algorithm) && !"SHA256withRSAandMGF1".equals(algorithm) && !"SHA256withIso9796d2PSS".equals(algorithm)) {
            throw new NoSuchAlgorithmException(algorithm);
        }
        this.algorithm = algorithm;
    }

    @Override
    protected void engineInitVerify(PublicKey publicKey) throws InvalidKeyException {
        throw new UnsupportedOperationException();
    }

    @Override
    protected void engineInitSign(PrivateKey privateKey) throws InvalidKeyException {
        if (privateKey instanceof CardBasedRSAPrivateKey) {
            CardBasedRSAPrivateKey cardBasedKey = (CardBasedRSAPrivateKey) privateKey;
            this.card = cardBasedKey.getCard();
            this.dfSpecificPrivateKeyRef = cardBasedKey.getDfSpecificPrivateKeyRef();
        } else {
            throw new InvalidKeyException("Expected key of type CardBasedRSAPrivateKey but received key of type " + privateKey.getClass());
        }
    }

    @Override
    protected void engineUpdate(byte b) throws SignatureException {
        // TODO b should be appended to input instead of replacing it
        input = new byte[] {b};
    }

    @Override
    protected void engineUpdate(byte[] b, int off, int len) throws SignatureException {
        // TODO b should be appended to input instead of replacing it
        input = Arrays.copyOfRange(b, off, off + len);
    }

    @Override
    protected byte[] engineSign() throws SignatureException {
        try {
            if (algorithm.equals("SHA256withRSA")) {
                return card.doRSASSA_PKCS1_V1_5_SIGN(input, dfSpecificPrivateKeyRef);
            } else if (algorithm.equals("SHA256withRSAandMGF1")) {
                return card.doPSSSigner(input, dfSpecificPrivateKeyRef);
            } else if (algorithm.equals("SHA256withIso9796d2PSS")) {
                return card.doIso9796d2PSSSigner(input, dfSpecificPrivateKeyRef);
            } else {
                throw new SignatureException("Unknown algorithm: " + algorithm);
            }
        } catch (CardException e) {
            throw new SignatureException(e);
        } catch (IOException e) {
            throw new SignatureException(e);
        } catch (Iso7816UnsuccessfulExecutionException e) {
            throw new SignatureException(e);
        } finally {
            // clean up everything
            dfSpecificPrivateKeyRef = -1;
            card = null;
            input = null;
        }
    }

    @Override
    protected boolean engineVerify(byte[] sigBytes) throws SignatureException {
        throw new UnsupportedOperationException();
    }

    @Override
    public void engineSetParameter(AlgorithmParameterSpec aps) {
        this.aps = aps;
    }

    @Override
    protected void engineSetParameter(String param, Object value) throws InvalidParameterException {
        parameters.put(param, value);
    }

    @Override
    protected Object engineGetParameter(String param) throws InvalidParameterException {
        return parameters.get(param);
    }


    /**
     * Convenience class for algorithm SHA256withRSA.
     */
    public static final class SHA256withRSA extends CardBasedSignature {
        public SHA256withRSA() throws NoSuchAlgorithmException {
            super("SHA256withRSA");
        }
    }

    /**
     * Convenience class for algorithm SHA256withRSAandMGF1.
     */
    public static final class SHA256withRSAandMGF1 extends CardBasedSignature {
        public SHA256withRSAandMGF1() throws NoSuchAlgorithmException {
            super("SHA256withRSAandMGF1");
        }
    }

    /**
     * Convenience class for algorithm SHA256withIso9796d2PSS.
     */
    public static final class SHA256withIso9796d2PSS extends CardBasedSignature {
        public SHA256withIso9796d2PSS() throws NoSuchAlgorithmException {
            super("SHA256withIso9796d2PSS");
        }
    }

}
