/**
 * Copyright 2011, Tobias Senger
 * <p/>
 * This file is part of animamea.
 * <p/>
 * Animamea is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * <p/>
 * Animamea is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU General Public License
 * along with animamea.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.gematik.security.tr03110;

import de.gematik.security.Utilities.Utilities;
import de.gematik.security.crypto.AmCryptoException;
import de.gematik.security.crypto.AmCryptoProvider;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.spongycastle.asn1.ASN1InputStream;

import javax.smartcardio.CommandAPDU;
import javax.smartcardio.ResponseAPDU;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

/**
 * Verpackt ungeschützte CAPDU in SecureMessaging und entpackt SM-geschützte RAPDU.
 *
 * @author Tobias Senger (tobias@t-senger.de)
 */
public class SecureMessaging {
    private static final Log LOG = LogFactory.getLog(SecureMessaging.class);

    private byte[] ks_enc = null;
    private byte[] ks_mac = null;
    private SSC ssc = null;
    private AmCryptoProvider crypto = null;

    /**
     * Konstruktor
     *
     * @param acp        AmDESCrypto- oder AmAESCrypto-Instanz
     * @param ksenc      Session Key für Verschlüsselung (K_enc)
     * @param ksmac      Session Key für Prüfsummenberechnung (K_mac)
     * @param initialSSC Initialer Wert des Send Sequence Counters
     */
    public SecureMessaging(AmCryptoProvider acp, byte[] ksenc, byte[] ksmac,
                           byte[] initialSSC) {

        crypto = acp;

        ks_enc = ksenc.clone();
        ks_mac = ksmac.clone();

        ssc = new SSC(initialSSC.clone());
    }

    /**
     * Erzeugt aus einer Plain-Command-APDU ohne Secure Messaging eine Command-APDU
     * mit Secure Messaging.
     *
     * @param commandAPDU plain CommandAPDU
     * @return protected CommandAPDU
     * @throws SecureMessagingException
     * @throws IllegalArgumentException
     */
    public CommandAPDU wrap(CommandAPDU commandAPDU) throws SecureMessagingException {

        byte[] header = null;
        int lc = 0;
        boolean isExtended;
        int apduCase;
        DO97 do97 = null;
        DO87 do87 = null;
        DO8E do8E = null;

        ssc.increment();
        LOG.info("SCC: " + Utilities.ByteArrayToSpacedHexString(ssc.getValue()));

        //parse Command APDU
        /**
         * Command APDU encoding options:
         *
         * case 1:  |CLA|INS|P1 |P2 |                                 len = 4
         * case 2s: |CLA|INS|P1 |P2 |LE |                             len = 5
         * case 3s: |CLA|INS|P1 |P2 |LC |...BODY...|                  len = 6..260
         * case 4s: |CLA|INS|P1 |P2 |LC |...BODY...|LE |              len = 7..261
         * case 2e: |CLA|INS|P1 |P2 |00 |LE1|LE2|                     len = 7
         * case 3e: |CLA|INS|P1 |P2 |00 |LC1|LC2|...BODY...|          len = 8..65542
         * case 4e: |CLA|INS|P1 |P2 |00 |LC1|LC2|...BODY...|LE1|LE2|  len =10..65544
         *
         * LE, LE1, LE2 may be 0x00.
         * LC must not be 0x00 and LC1|LC2 must not be 0x00|0x00
         */
        byte[] apdu = commandAPDU.getBytes();
        if (apdu.length < 4) {
            throw new IllegalArgumentException("apdu must be at least 4 bytes long");
        }
        if (apdu.length == 4) {
            apduCase = 1;
            isExtended = false;
        } else {
            int l1 = apdu[4] & 0xff;
            if (l1 != 0) {
                isExtended = false;
                if (apdu.length == 5) {
                    apduCase = 2;
                } else if (apdu.length == 4 + 1 + l1) {
                    apduCase = 3;
                } else if (apdu.length == 4 + 2 + l1) {
                    apduCase = 4;
                } else {
                    throw new IllegalArgumentException
                            ("Invalid APDU: length=" + apdu.length + ", LC=" + l1);
                }
            } else {
                if(apdu.length==5){
                    isExtended = false;
                    apduCase = 2;
                }else if(apdu.length == 6) {
                    throw new IllegalArgumentException("extended apdu must be at least 7 bytes long");
                }else{
                    isExtended = true;
                    if (apdu.length == 7) {
                        apduCase = 2;
                    }else {
                        int l2 = ((apdu[5] & 0xff) << 8) | (apdu[6] & 0xff);
                        if (apdu.length == 4 + 3 + l2) {
                            apduCase = 3;
                        } else if (apdu.length == 4 + 5 + l2) {
                            apduCase = 4;
                        } else {
                            throw new IllegalArgumentException("Invalid APDU: length="
                                    + apdu.length + ", LC=" + l2);
                        }
                    }
                }
            }
        }

        // build header
        header = new byte[4];
        System.arraycopy(commandAPDU.getBytes(), 0, header, 0, 4);

        // set secure messaging bits in CLA-Byte
        header[0] = (byte) (header[0] | (byte) 0x0C);

        // build DO87
        if (apduCase == 3 || apduCase == 4) {
            do87 = buildDO87(commandAPDU.getData().clone());
            try {
                lc += do87.getEncoded().length;
            } catch (IOException e) {
                throw new SecureMessagingException(e);
            }
        }

        // build DO97
        if (apduCase == 2 || apduCase == 4) {
            do97 = buildDO97(commandAPDU.getNe());
            lc += do97.getEncoded().length;
        }

        // build DO8E
        do8E = buildDO8E(header, do87, do97);

        lc += do8E.getEncoded().length;

        // construct and return protected APDU
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();
        try

        {
            bOut.write(header);
            if (lc>255 || isExtended) {
                bOut.write(0);
                bOut.write(lc >> 8);
            }
            bOut.write(lc);
            if (do87 != null)
                bOut.write(do87.getEncoded());
            if (do97 != null)
                bOut.write(do97.getEncoded());
            bOut.write(do8E.getEncoded());
            if (lc>255 || isExtended) {
                bOut.write(0);
            }
            bOut.write(0);
        } catch (IOException e) {
            throw new SecureMessagingException(e);
        }

        return new CommandAPDU(bOut.toByteArray());
    }

    /**
     * Erzeugt aus einer SM geschützten Response-APDU eine plain Response-APDU
     * ohne Secure Messaging.
     *
     * @param rapdu SM protected RAPDU
     * @return plain RAPDU
     * @throws SecureMessagingException
     */
    public ResponseAPDU unwrap(ResponseAPDU rapdu) throws SecureMessagingException {

        DO81 do81 = null;
        DO87 do87 = null;
        DO99 do99 = null;
        DO8E do8E = null;

        ssc.increment();
        LOG.info("SCC: " + Utilities.ByteArrayToSpacedHexString(ssc.getValue()));

        int pointer = 0;
        byte[] rapduBytes = rapdu.getData();
        byte[] subArray = new byte[rapduBytes.length];

        while (pointer < rapduBytes.length) {
            System.arraycopy(rapduBytes, pointer, subArray, 0,
                    rapduBytes.length - pointer);
            ASN1InputStream asn1sp = new ASN1InputStream(subArray);
            byte[] encodedBytes = null;
            try {
                encodedBytes = asn1sp.readObject().getEncoded();
                asn1sp.close();
            } catch (IOException e) {
                throw new SecureMessagingException(e);
            }

            ASN1InputStream asn1in = new ASN1InputStream(encodedBytes);
            try {
                switch (encodedBytes[0]) {
                    case (byte) 0x81:
                        do81 = new DO81();
                        do81.fromByteArray(asn1in.readObject().getEncoded());
                        break;
                    case (byte) 0x87:
                        do87 = new DO87();
                        do87.fromByteArray(asn1in.readObject().getEncoded());
                        break;
                    case (byte) 0x99:
                        do99 = new DO99();
                        do99.fromByteArray(asn1in.readObject().getEncoded());
                        break;
                    case (byte) 0x8E:
                        do8E = new DO8E();
                        do8E.fromByteArray(asn1in.readObject().getEncoded());
                }
                asn1in.close();
            } catch (IOException e) {
                throw new SecureMessagingException(e);
            }

            pointer += encodedBytes.length;
        }

        if (do99 == null)
            throw new SecureMessagingException("Secure Messaging error: mandatory DO99 not found"); // DO99 is mandatory
        // and only absent
        // if SM error
        // occurs

        // Construct K (SSC||DO87||DO99)
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        try {
            if (do87 != null) {
                bout.write(do87.getEncoded());
            } else if (do81 != null) {
                bout.write(do81.getEncoded());
            }
            bout.write(do99.getEncoded());
        } catch (IOException e) {
            throw new SecureMessagingException(e);
        }

        crypto.init(ks_mac, ssc.getValue());
        byte[] cc = crypto.getMAC(bout.toByteArray());

        byte[] do8eData = do8E.getData();

        if (!java.util.Arrays.equals(cc, do8eData))
            throw new SecureMessagingException("Checksum is incorrect!\n Calculated CC: "
                    + Utilities.ByteArrayToSpacedHexString(cc) + "\nCC in DO8E: "
                    + Utilities.ByteArrayToSpacedHexString(do8eData));

        // Decrypt DO87
        byte[] data = null;
        byte[] unwrappedAPDUBytes = null;

        if (do87 != null) {
            crypto.init(ks_enc, ssc.getValue());
            byte[] do87Data = do87.getData();
            try {
                data = crypto.decrypt(do87Data);
            } catch (AmCryptoException e) {
                throw new SecureMessagingException(e);
            }
        } else if (do81 != null) {
            data = do81.getData();
        }
        // Build unwrapped RAPDU
        if (data != null) {
            unwrappedAPDUBytes = new byte[data.length + 2];
            System.arraycopy(data, 0, unwrappedAPDUBytes, 0, data.length);
            byte[] do99Data = do99.getData();
            System.arraycopy(do99Data, 0, unwrappedAPDUBytes, data.length,
                    do99Data.length);
        } else {
            unwrappedAPDUBytes = do99.getData().clone();
        }

        return new ResponseAPDU(unwrappedAPDUBytes);
    }


    /**
     * encrypt data with KS.ENC and build DO87
     *
     * @param data
     * @return
     * @throws SecureMessagingException
     */
    private DO87 buildDO87(byte[] data) throws SecureMessagingException {

        crypto.init(ks_enc, ssc.getValue());
        byte[] enc_data;
        try {
            enc_data = crypto.encrypt(data);
        } catch (AmCryptoException e) {
            throw new SecureMessagingException(e);
        }

        return new DO87(enc_data);

    }

    private DO8E buildDO8E(byte[] header, DO87 do87, DO97 do97) throws SecureMessagingException {

        ByteArrayOutputStream m = new ByteArrayOutputStream();

        /**
         * Verhindert doppeltes Padden des Headers: Nur wenn do87 oder do97
         * vorhanden sind, wird der Header gepaddet. Ansonsten wird erst beim
         * Berechnen des MAC gepaddet.
         */
        try {
            if (do87 != null || do97 != null)
                m.write(crypto.addPadding(header));

            else
                m.write(header);

            if (do87 != null)
                m.write(do87.getEncoded());
            if (do97 != null)
                m.write(do97.getEncoded());
        } catch (IOException e) {
            throw new SecureMessagingException(e);
        }

        crypto.init(ks_mac, ssc.getValue());

        return new DO8E(crypto.getMAC(m.toByteArray()));
    }

    private DO97 buildDO97(int le) {
        return new DO97(le);
    }

    private static class SSC{
        private byte[] value;
        public SSC(byte[] initialValue){
            value = initialValue;
        }

        public void increment(){
            increment(value.length-1);
        }

        private void increment(int index){
            value[index]++;
            if(value[index]==0 && index>0){
                increment(index - 1);
            }
        }

        public byte[] getValue() {
            return value;
        }
    }
}
