/*
 * HealthcardIO - a library providing functionality for accessing healthcards
 *
 * Copyright (C) 2017  gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package de.gematik.security.test;

import de.gematik.security.Utilities.CardFileSystem;
import de.gematik.security.Utilities.CardItemFile;
import de.gematik.security.Utilities.Utilities;
import de.gematik.security.impl.HealthcardContext;
import de.gematik.security.iso7816.Constants;
import de.gematik.security.iso7816.FCP;
import de.gematik.security.iso7816.Iso7816UnsuccessfulExecutionException;
import de.gematik.security.iso9564.PinBlockException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.json.JSONException;
import org.junit.Test;
import org.xml.sax.SAXException;

import javax.smartcardio.CardException;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ShortBuffer;
import java.security.cert.CertificateException;
import java.util.Iterator;

import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

/**
 * @author André Szdzuy
 */
public class NfcHealthcardScenarioTest extends NfcBaseTest {
    private static final Log LOG = LogFactory.getLog(NfcHealthcardScenarioTest.class);

    @Test
    public void readVSD() throws Iso7816UnsuccessfulExecutionException, CardException, IOException, PinBlockException {
        LOG.info("readVSD()");
        assertNotNull("healthcardContext is null", healthcardContext);
        assertTrue("The inserted card is no EGK, this test is only useful for EGKs.", healthcardContext.isEGK());

        //SELECT DF.HCA
        healthcardContext.selectADF("D27600000102");

        //SELECT EF.PD
        healthcardContext.selectEF("D001");

        //Read length
        ShortBuffer shortBuffer = ByteBuffer.wrap(healthcardContext.read(0, 2)).asShortBuffer();

        //Read "Persönliche" VSD
        byte[] data = healthcardContext.read(2, shortBuffer.get(0));
        String result = Utilities.unzip(data);
        LOG.info(result);

//        byte[] data2 = Utilities.zip(result);
//        assertArrayEquals(data, data2);

        //SELECT EF.VD
        healthcardContext.selectEF("D002");

        //Read offsets
        shortBuffer = ByteBuffer.wrap(healthcardContext.read(0, 8)).asShortBuffer();

        //Read "Allgemeine" VSD
        data = healthcardContext.read(shortBuffer.get(0), shortBuffer.get(1) - shortBuffer.get(0) + 1);
        result = Utilities.unzip(data);
        LOG.info(result);

        //SELECT EF.GVD
        healthcardContext.selectEF("D003");

        //Read length
        shortBuffer = ByteBuffer.wrap(healthcardContext.read(0, 2)).asShortBuffer();

        //Read "Geschütze" VSD
        data = healthcardContext.read(2, shortBuffer.get(0));
        result = Utilities.unzip(data);
        LOG.info(result);
    }

    @Test
    public void exploreCard() throws Iso7816UnsuccessfulExecutionException, CardException, CertificateException, JSONException, IOException, PinBlockException, ParserConfigurationException, SAXException {
        LOG.info("exploreCard()");
        assertNotNull("healthcardContext is null", healthcardContext);

        FCP fcp = healthcardContext.selectADFAndReturnFCP();
        LOG.info("Root application identifier: " + Utilities.ByteArrayToSpacedHexString(fcp.getDfName()));
        exploreEFs(cfs, healthcardContext, fcp);

        Iterator<String> aidIterator = cfs.getAllAids();
        boolean isNextAidSupported = true;
        while (true) {
            if (isNextAidSupported) {
                try {
                    fcp = healthcardContext.selectADFAndReturnFCP(true);
                } catch (Iso7816UnsuccessfulExecutionException e) {
                    if (e.getSW() == 0x6a86) {
                        isNextAidSupported = false;
                        healthcardContext.verify("123456");
                        continue;
                    } else if (e.getSW() == Constants.APDU_RESPONSE_FILE_NOT_FOUND) {
                        break;
                    } else {
                        throw e;
                    }
                }
            } else {
                if (aidIterator.hasNext()) {
                    try {
                        String aid = aidIterator.next();
                        if ("D2760001448000".equals(aid) || "D27600014601".equals(aid) || "D27600014606".equals(aid)) { //skip MF
                            continue;
                        }
                        fcp = healthcardContext.selectADFAndReturnFCP(aid);
                    } catch (Iso7816UnsuccessfulExecutionException e) {
                        if (e.getSW() == Constants.APDU_RESPONSE_FILE_NOT_FOUND) {
                            continue;
                        }
                        throw e;
                    }
                } else {
                    break;
                }
            }
            String aid = Utilities.ByteArrayToHexString(fcp.getDfName());
            String name = cfs.getDF(aid);
            LOG.info("ADF application identifier: " + aid + " = " + name);
            if (!name.equals("DF.OSE")) {
                // skip DF.OSE because after error 0x6F00 occured (see comment in method exploreEFs()) the card is in
                // a strange state that does not allow continuing, so we prevent error 0x6F00 by skipping DF.OSE
                exploreEFs(cfs, healthcardContext, fcp);
            }
        }
    }

    private void exploreEFs(CardFileSystem cfs, HealthcardContext channel, FCP parent) throws Iso7816UnsuccessfulExecutionException, CardException, CertificateException, IOException, JSONException, ParserConfigurationException, SAXException {
        Iterator<String> fidIterator = cfs.getFids(Utilities.ByteArrayToHexString(parent.getDfName()));

        while (fidIterator.hasNext()) {
            String fid = fidIterator.next();
            try {
                CardItemFile cardItemFile = channel.readEF(cfs, parent, fid);
                LOG.info("Data of file " + cardItemFile.getName() + ":");
                LOG.info(cardItemFile.displayData(false));
            } catch (Iso7816UnsuccessfulExecutionException e) {
                if(e.getSW() == Constants.APDU_RESPONSE_SECURITY_STATUS_NOT_SATISFIED || e.getSW() == 0x6F00) {
                    // 0x6F00 is an unknown error code that only occurs with file DF.OSE and only in non-NFC mode
                    // on the Windows computer, this error code also has something to do with the LE value of
                    // read operations (see comment in class Card2CardTest)
                    LOG.error("Security status not satisfied for file: fileIdentifier=" + fid);
                } else {
                    throw e;
                }
            }
        }
    }

}
